/* This file contains the implementation of routines private to this device */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/root.h>
#include <sys/time.h>
#include "mpid.h"

LinkCB_t      **SendLinkList = NULL;
LinkCB_t      **RecvLinkList = NULL;
Option_t       *RecvOptionList = NULL;
Thread_t      **SendThreadList = NULL;

#define PARIX_OK 0
#define PARIX_ERROR -1
#define STDERR stdout
#define DEBUG(a) {a}
#define SEND_STACKSIZE 4096

typedef struct {
  LinkCB_t       *link;
  char           *buf;
  int             size;
  int             dest;
  Thread_t       *Tptr;
  int            *cmpl;
}               Par_SendLink_t;

int             Wait_SendLink(Thread_t *Tptr,int partner);
int             Proc_SendLink(Par_SendLink_t * slpar);

static int MPID_PARIX_Debugflag=0;

void MPID_PARIX_SetDebugFlag( void *ctx, int f )
{
  MPID_PARIX_Debugflag = f;
}

void mpid_parix_setdebugflag(void *ctx, int *f)
{
  MPID_PARIX_Debugflag = *f;
}

/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_RecvFromChannel
 *  ARGS:
 *    O   buf          Buffer for incoming data
 *    I   maxsize      Size of buffer supplied
 *    I   from         Which process to receive the data from
 *  RETURNS: void
 *
 * Read a data block from a process
 *
\*----------------------------------------------------------------------*/

void            MPID_PARIX_RecvFromChannel(void *buf, int maxsize, int from)
{
  int             result = 0;
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) 
  DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] RecvChannel from %d size %d\n", MPID_MyWorldRank, from, maxsize);)
#endif				/* #DEBUG_END# */
    if (from < 0 || from > MPID_WorldSize - 1) {
    fprintf(STDERR, "[%d]IP ERROR from %d out of range (%s:%d)\n",
	    MPID_MyWorldRank, from, __FILE__, __LINE__);
    MPID_PARIX_Abort(-1);
  }
  if (-1 == (result = RecvLink(RecvLinkList[from], buf, maxsize))) {
    switch (result=errno) {
    case EINVAL:
      fprintf(STDERR, "[%d]R ERROR RecvLink EINVAL\n", MPID_MyWorldRank);
      MPID_PARIX_Abort(-1);
      break;
    case EMSGSIZE:
      fprintf(STDERR, "[%d]R ERROR RecvLink EMSGSIZE\n", MPID_MyWorldRank);
      fflush(STDERR);
      break;
    default:
      fprintf(STDERR, "[%d]R ERROR RecvLink UNKNOWN\n", MPID_MyWorldRank);
      MPID_PARIX_Abort(-1);
      break;
    }
    result=-result;
  }
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) 
  DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] RecvChannel from %d complete %d\n", MPID_MyWorldRank, from, result);)
#endif				/* #DEBUG_END# */
    return;
}

/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_RecvAnyControl
 *  ARGS:
 *    O   pkt       Buffer in which to receive control message
 *    I   size      Size of buffer provided, not to be exceeded.
 *    O   from      Which process the control message is from.
 *  RETURNS: void
 *
 * This function returns a control message. It will block until one
 * becomes available.
 *
\*----------------------------------------------------------------------*/


void            MPID_PARIX_RecvAnyControl(MPID_PKT_T * pkt, int size, int *from)
{
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) 
  DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] RecvAnyControl\n", MPID_MyWorldRank);)
#endif				/* #DEBUG_END# */

    *from = SelectList(MPID_WorldSize, RecvOptionList);
  MPID_PARIX_RecvFromChannel((void *) pkt, size, *from);
  return;
}



/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_ControlMsgAvail
 *  ARGS: void
 *
 *  RETURNS: int
 *
 * Returns TRUE if a control message is available, FALSE otherwise.
 *
\*----------------------------------------------------------------------*/

int             MPID_PARIX_ControlMsgAvail(void)
{
  int result=CondSelectList(MPID_WorldSize, RecvOptionList);
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) 
  DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] ControlMsgAvail %d\n",   
    MPID_MyWorldRank,result);)
#endif				/* #DEBUG_END# */
  return (-1==result ? FALSE : TRUE);
}

/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_SendControlBlock
 *  ARGS:
 *    I   pkt       Control message to be sent
 *    I   size      Size of control message
 *    I   dest      Who to send message to
 *  RETURNS: void
 *
 * Send a control message to another process, allows the send to wait
 * until the message is received (but does NOT require it).
 * This can simplify some buffer handling.
 *
\*----------------------------------------------------------------------*/

void            MPID_PARIX_SendControlBlock(MPID_PKT_T * pkt, int size, int to)
{ int result;
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) 
  DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] SendControlBlock to %d size %d\n", MPID_MyWorldRank, to, size);)
#endif				/* #DEBUG_END# */

    if (to < 0 || to > MPID_WorldSize - 1) {
    fprintf(STDERR, "[%d]S ERROR to %d out of range (%s:%d)\n",
	    MPID_MyWorldRank, to, __FILE__, __LINE__);
    MPID_PARIX_Abort(-1);
  }
  Wait_SendLink(SendThreadList[to],to);
  SendThreadList[to]=NULL;

  if (-1 == (result = SendLink(SendLinkList[to], pkt, size))) {
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
    switch (errno) {
    case EINVAL:
      fprintf(STDERR, "[%d]S ERROR SendControlBlock EINVAL\n", MPID_MyWorldRank);
      MPID_PARIX_Abort(-1);
      break;
    case EMSGSIZE:
      fprintf(STDERR, "[%d]S ERROR SendControlBlock EMSGSIZE\n", MPID_MyWorldRank);
      fflush(STDERR);
      break;
    default:
      fprintf(STDERR, "[%d]S ERROR SendControlBlock UNKNOWN\n", MPID_MyWorldRank);
      MPID_PARIX_Abort(-1);
      break;
    }
#endif				/* #DEBUG_END# */
  }
  return;
}

/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_SendControl
 *  ARGS:
 *    I   pkt       Control message to be sent
 *    I   size      Size of control message
 *    I   dest      Who to send message to
 *  RETURNS: void
 *
 * Send a control message to another process
 *
\*----------------------------------------------------------------------*/

void            MPID_PARIX_SendControl(MPID_PKT_T * pkt, int size, int to)
{
  int             result, error;
  Par_SendLink_t *Par_SendLink = (Par_SendLink_t *)
  malloc(sizeof(Par_SendLink_t));
  if (!Par_SendLink) {
    fprintf(STDERR, "[%d]IS out of memory (%s:%d)\n",
	    MPID_MyWorldRank, __FILE__, __LINE__);
    MPID_PARIX_Abort(-1);
  }
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag)
    DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] SendControl %d to %d size %d (%s:%d)\n", MPID_MyWorldRank, pkt, to, size, __FILE__, __LINE__);)
#endif				/* #DEBUG_END# */
    if (to < 0 || to > MPID_WorldSize - 1) {
    fprintf(STDERR, "[%d]S ERROR to %d out of range (%s:%d)\n",
	    MPID_MyWorldRank, to, __FILE__, __LINE__);
    MPID_PARIX_Abort(-1);
  }
  Par_SendLink->link = SendLinkList[to];
  /* short packets are always buffered */
#ifdef FOO
  if (size > sizeof(MPID_PKT_T))
    Par_SendLink->buf = (char *) pkt;
  else
#endif
  {
    Par_SendLink->buf = malloc(size);
    if (!Par_SendLink->buf) {
      fprintf(STDERR, "[%d]IS out of memory (%s:%d)\n",
	      MPID_MyWorldRank, __FILE__, __LINE__);
      MPID_PARIX_Abort(-1);
    }
    MEMCPY(Par_SendLink->buf, (char *) pkt, size);
  }
  Par_SendLink->size = size;
  Par_SendLink->dest = to;
  Par_SendLink->Tptr = SendThreadList[to];
  Par_SendLink->cmpl = NULL;
  SendThreadList[to] = NULL;
  if (NULL == (SendThreadList[to] =
#ifdef PARIX_1_2_T8
	       StartThread(Proc_SendLink, SEND_STACKSIZE, &error,
			   4, (void *) Par_SendLink))) {
#else
	       CreateThread(NULL, SEND_STACKSIZE,
			    Proc_SendLink, &error, (void *) Par_SendLink))) {
#endif
    switch (error) {
    case ENOMEM:
      fprintf(STDERR, "[%d]IS out of memory, can't create thread (%s:%d)\n",
	      MPID_MyWorldRank, __FILE__, __LINE__);
      MPID_PARIX_Abort(-1);
      break;
    default:
      fprintf(STDERR, "[%d]IS can't create thread (%s:%d)\n",
	      MPID_MyWorldRank, __FILE__, __LINE__);
      MPID_PARIX_Abort(-1);
      break;
    }
  }
  return;
}


/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_SendChannel
 *  ARGS:
 *    I   buf          Buffer containing data to be sent
 *    I   size         Size of buffer
 *    I   dest         Who to send the data to
 *  RETURNS: void
 *
 * Send contents of supplied buffer to another process
 *
\*----------------------------------------------------------------------*/

void            MPID_PARIX_SendChannel(void *pkt, int size, int to)
{
  int             result;
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) 
  DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] SendChannel to %d size %d (%s:%d)\n", MPID_MyWorldRank, to, size,__FILE__, __LINE__);)
#endif				/* #DEBUG_END# */
    if (to < 0 || to > MPID_WorldSize - 1) {
    fprintf(STDERR, "[%d]S ERROR to %d out of range (%s:%d)\n",
	    MPID_MyWorldRank, to, __FILE__, __LINE__);
    MPID_PARIX_Abort(-1);
  }
  Wait_SendLink(SendThreadList[to],to);
  SendThreadList[to]=NULL;
  if (-1 == (result = SendLink(SendLinkList[to], pkt, size))) {
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
    switch (result=errno) {
    case EINVAL:
      fprintf(STDERR, "[%d]S ERROR SendChannel EINVAL\n", MPID_MyWorldRank);
      MPID_PARIX_Abort(-1);
      break;
    case EMSGSIZE:
      fprintf(STDERR, "[%d]S ERROR SendChannel EMSGSIZE\n", MPID_MyWorldRank);
      fflush(STDERR);
      break;
    default:
      fprintf(STDERR, "[%d]S ERROR SendChannel UNKNOWN\n", MPID_MyWorldRank);
      MPID_PARIX_Abort(-1);
      break;
    }
    result=-result;
#endif				/* #DEBUG_END# */
  }
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) 
  DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] SendChannel to %d complete %d\n", MPID_MyWorldRank, to, result);)
#endif				/* #DEBUG_END# */
  return;
}


/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_ISendChannel
 *  ARGS:
 *    I   buf          Buffer containing data to be sent
 *    I   size         Size of buffer
 *    I   dest         Who to send the data to
 *    O   sid	       Id
 *  RETURNS: void
 *
 * Send contents of supplied buffer to another process
 *
\*----------------------------------------------------------------------*/

void            MPID_PARIX_ISendChannel(void *pkt, int size, int to, int *sid)
{
  int             result, error;
  Par_SendLink_t *Par_SendLink = (Par_SendLink_t *)
  malloc(sizeof(Par_SendLink_t));
  if (!Par_SendLink) {
    fprintf(STDERR, "[%d]IS out of memory (%s:%d)\n",
	    MPID_MyWorldRank, __FILE__, __LINE__);
    MPID_PARIX_Abort(-1);
  }
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag)
    DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] ISendChannel %d to %d size %d (%s:%d)\n", MPID_MyWorldRank, pkt, to, size, __FILE__, __LINE__);)
#endif				/* #DEBUG_END# */
    if (to < 0 || to > MPID_WorldSize - 1) {
    fprintf(STDERR, "[%d]S ERROR to %d out of range (%s:%d)\n",
	    MPID_MyWorldRank, to, __FILE__, __LINE__);
    MPID_PARIX_Abort(-1);
  }
  Par_SendLink->link = SendLinkList[to];
  Par_SendLink->buf = (char *) pkt;
  Par_SendLink->size = size;
  Par_SendLink->dest = to;
  Par_SendLink->Tptr = SendThreadList[to];
  Par_SendLink->cmpl = sid;
  SendThreadList[to] = NULL;
  if (NULL == (SendThreadList[to] =
#ifdef PARIX_1_2_T8
	       StartThread(Proc_SendLink, SEND_STACKSIZE, &error,
			   4, (void *) Par_SendLink))) {
#else
	       CreateThread(NULL, SEND_STACKSIZE,
			    Proc_SendLink, &error, (void *) Par_SendLink))) {
#endif
    switch (error) {
    case ENOMEM:
      fprintf(STDERR, "[%d]IS out of memory, can't create thread (%s:%d)\n",
	      MPID_MyWorldRank, __FILE__, __LINE__);
      MPID_PARIX_Abort(-1);
      break;
    default:
      fprintf(STDERR, "[%d]IS can't create thread (%s:%d)\n",
	      MPID_MyWorldRank, __FILE__, __LINE__);
      MPID_PARIX_Abort(-1);
      break;
    }
  }
  return;
}


/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_WSendChannel
 *  ARGS:
 *    I   buf          Buffer containing data to be sent
 *    I   size         Size of buffer
 *    I   dest         Who to send the data to
 *    I   sid	       Id
 *  RETURNS: void
 *
 * Wait for completion of a nonblocking send of data
 *
\*----------------------------------------------------------------------*/
#define WTIME 1000 
/* CLOCK_TICK=WTIME*WTIME */

void            MPID_PARIX_WSendChannel(void *pkt, int size, int to, int *sid)
{
  int             count = 0;
  while (*sid) {
    TimeWait(WTIME);
    count++;
    if (count % WTIME)
      fprintf(STDERR, "[%d] WSendChannel waits %f sec\n",
	      MPID_MyWorldRank,(double)count/WTIME);
  }
  return;
}

/*----------------------------------------------------------------------*\
 * FUNCTION: Wait_SendLink
 *  ARGS:
 *    I   Tptr         thread waited for
 *    I   partner      ID of Send dest
 *  RETURNS: void
 *
 * Wait for completion of a specified Thread
 *
\*----------------------------------------------------------------------*/
int             Wait_SendLink(Thread_t *Tptr, int partner)
{

  int             result, error;

  if (Tptr) {
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) {
     DEBUG(fprintf(MPID_DEBUG_FILE,  
           "[%d]Wait_SendLink to %d handle %d (%s:%d)\n",
	    MPID_MyWorldRank, partner, Tptr,
            __FILE__, __LINE__);)
  }
#endif				/* #DEBUG_END# */
    WaitThread(Tptr, &result);
    /* result is return value of SendLink */
    /* errno is return value of WaitThread */
    if (error=errno) {
      fprintf(STDERR, "[%d]IS ERROR %d Termination Thread %d to %d\n",
          MPID_MyWorldRank, error, Tptr,partner);
    /*  MPID_PARIX_Abort(-1); */
    }
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) 
  DEBUG(fprintf(MPID_DEBUG_FILE,
        "[%d] Wait_Thread %d to %d complete %d\n", 
         MPID_MyWorldRank, Tptr, partner, result);)
#endif				/* #DEBUG_END# */
  }
  return PARIX_OK;
}

/*----------------------------------------------------------------------*\
 * FUNCTION: Proc_SendLink
 *  ARGS:
 *    I   slpar         struct with Send-pars
 *  RETURNS: void
 *
 * Wait for completion of a specified Thread
 *
\*----------------------------------------------------------------------*/
int             Proc_SendLink(Par_SendLink_t * slpar)
{
  int             result = 0;
  /* parameters already checked */
  Wait_SendLink(slpar->Tptr,slpar->dest);
  if (-1 == (result = SendLink(slpar->link,
			       slpar->buf, slpar->size))) {
    switch (result=errno) {
    case EINVAL:
      fprintf(STDERR, "[%d]IS ERROR SendLink EINVAL\n", MPID_MyWorldRank);
      MPID_PARIX_Abort(-1);
      break;
    case EMSGSIZE:
      fprintf(STDERR, "[%d]IS ERROR SendLink EMSGSIZE\n", MPID_MyWorldRank);
      fflush(STDERR);
      break;
    default:
      fprintf(STDERR, "[%d]IS ERROR SendLink UNKNOWN\n", MPID_MyWorldRank);
      MPID_PARIX_Abort(-1);
      break;
    }
    result=-result;
  }
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
  if (MPID_PARIX_Debugflag) 
  DEBUG(fprintf(MPID_DEBUG_FILE,
        "[%d] Proc_SendLink to %d complete %d\n", 
         MPID_MyWorldRank, slpar->dest, result);)
#endif				/* #DEBUG_END# */
  if (slpar->cmpl)
    *(slpar->cmpl)=0;
  if (slpar->size <= sizeof(MPID_PKT_T))
    free(slpar->buf);
  free(slpar);
  return result;
}

/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_PARIX_init
 *  ARGS:
 *    I   argc
 *    I   argv
 *  RETURNS: void
 *
 * Initialize MPI, must be done only once (should be checked)
 *
\*----------------------------------------------------------------------*/
void            MPID_PARIX_init(int *argc, char ***argv)
{
  char           *FName = "ppc_init";
  int             partner, error = 0;
  LinkCB_t       *llink[2];

#ifdef DEBUG_PARIX		/* #DEBUG_START# */
if (MPID_PARIX_Debugflag)
  DEBUG(fprintf(MPID_DEBUG_FILE, 
"[%d] init with %d procs\n", MPID_MyWorldRank, MPID_WorldSize);)
#endif				/* #DEBUG_END# */

  SendLinkList = (LinkCB_t **) malloc(MPID_WorldSize * sizeof(LinkCB_t *));
  RecvLinkList = (LinkCB_t **) malloc(MPID_WorldSize * sizeof(LinkCB_t *));
  RecvOptionList = (Option_t *) malloc(MPID_WorldSize * sizeof(Option_t));
  SendThreadList = (Thread_t **) malloc(MPID_WorldSize * sizeof(Thread_t *));
  if (SendLinkList == NULL || RecvLinkList == NULL ||
      RecvOptionList == NULL || SendThreadList == NULL) {
    fprintf(STDERR, "%s Speicherallokierung", FName);
    MPID_PARIX_Abort(1);
  }
  if (LocalLink(llink)) {
    LogError(EC_ERROR, FName, "LocalLink");
    AbortServer(1);
  }
  SendLinkList[MPID_MyWorldRank] = llink[0];
  RecvLinkList[MPID_MyWorldRank] = llink[1];
  RecvOptionList[MPID_MyWorldRank] =
    ReceiveOption(RecvLinkList[MPID_MyWorldRank]);

  for (partner = 0; partner < MPID_WorldSize; partner++) {
    SendThreadList[partner] = NULL;
    if (partner < MPID_MyWorldRank) {
      SendLinkList[partner] = MakeLink(partner, 0, &error);
      if (error) {
	fprintf(STDERR, "%s MakeLink", FName);
	MPID_PARIX_Abort(-1);
      }
      RecvLinkList[partner] = GetLink(partner, 1, &error);
      if (error) {
	fprintf(STDERR, "%s GetLink", FName);
	MPID_PARIX_Abort(-1);
      }
      RecvOptionList[partner] = ReceiveOption(RecvLinkList[partner]);
    } else if (partner > MPID_MyWorldRank) {
      RecvLinkList[partner] = GetLink(partner, 0, &error);
      if (error) {
	fprintf(MPID_DEBUG_FILE, "%s GetLink", FName);
	MPID_PARIX_Abort(-1);
      }
      RecvOptionList[partner] = ReceiveOption(RecvLinkList[partner]);
      SendLinkList[partner] = MakeLink(partner, 1, &error);
      if (error) {
	fprintf(MPID_DEBUG_FILE, "%s MakeLink", FName);
	MPID_PARIX_Abort(-1);
      }
    }
  }

#ifdef DEBUG_PARIX		/* #DEBUG_START# */
if (MPID_PARIX_Debugflag)
  DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] init complete\n", MPID_MyWorldRank);)
#endif

    return;
}

/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_PARIX_finalize
 *  ARGS:
 *  RETURNS: void
 *
 * Finalize MPI, waiting for all pending threads
 *
\*----------------------------------------------------------------------*/
void            MPID_PARIX_finalize(void)
{
  int             partner;
  for (partner = 0; partner < MPID_WorldSize; partner++)
    if (SendThreadList[partner]) {
#ifdef MPID_DEBUG_ALL		/* #DEBUG_START# */
      if (MPID_PARIX_Debugflag) {
	fprintf(MPID_DEBUG_FILE, "[%d]finalize Tptr to %d in use (%s:%d)\n",
		MPID_MyWorldRank, partner, __FILE__, __LINE__);
	fflush(STDERR);
      }
#endif				/* #DEBUG_END# */
      Wait_SendLink(SendThreadList[partner],partner);
      SendThreadList[partner]=NULL;
    }
  return;
}

/*----------------------------------------------------------------------*\
 * FUNCTION: MPID_PARIX_abort
 *  ARGS:
 *    I		text	Errormsg
 *    I		code	Errorcode
 *  RETURNS: void
 *
 * Shutdown MPI _and_ all parallel tasks
 *
\*----------------------------------------------------------------------*/
void            MPID_PARIX_abort(char *text, int code)
{
#ifdef DEBUG_PARIX		/* #DEBUG_START# */
if (MPID_PARIX_Debugflag)
  DEBUG(fprintf(MPID_DEBUG_FILE, "[%d] abort\n", MPID_MyWorldRank);)
#endif
    AbortServer(code);
}

/*----------------------------------------------------------------------*\
 * FUNCTION: getarg
 *  ARGS:
 *    I   k        number of arg
 *    O   arg      argv[k]
 *  RETURNS: void
 *
 *    The statement call getarg( k  ,  arg  )  will  get  the  kth
 *    command-line argument and put it into arg .
 *
 *    The 0th argument is the command name.
 *
 *    For unknown reasons this function is not part of the ppc-f77-lib
 *
\*----------------------------------------------------------------------*/



extern void     getarg(int *k, char *arg);

void            getarg(int *k, char *arg)
{
  int             i = *k;
  char          **argv = parix_environ->ArgV;
  while (--i) {
    *argv++;
    if (*argv == NULL)
      return;
  }
  strcpy(arg, *argv);
}

/*----------------------------------------------------------------------*\
 * FUNCTION: iargc
 *  ARGS:
 *  RETURNS: int   number of args
 *
 *    The function iargc returns the argument count: the index  of
 *    the last command-line argument,
 *    and therefore the number of arguments after the  command name
 *
 *    For unknown reasons this function is not part of the ppc-f77-lib
 *
\*----------------------------------------------------------------------*/
extern int      iargc(void);

int             iargc(void)
{
  char          **argv = parix_environ->ArgV;
  int             ret = 0;
  while (*++argv)
    ret++;
  return ret;
}

/*----------------------------------------------------------------------*\
 * FUNCTION: sleep
 *  ARGS:
 *    I		seconds
 *  RETURNS: unsigned 0
 *
 *    The function sleep dispatches the caller for seconds seconds
 *
 *    For unknown reasons this function is not part of the ppc-lib
 *
\*----------------------------------------------------------------------*/
unsigned        sleep(unsigned seconds)
{
  TimeWait(CLOCK_TICK * seconds);
  return 0;
}



char           *trmalloc(unsigned int size, int line, char *file)
{ char *ret;
  if (!(ret=malloc(size)))
    fprintf(stderr,"ERROR:MALLOC (%d:%s)",line,file);
  return ret;
}

char           *trcalloc(unsigned int number, int size, int line, char *file)
{ char *ret;
  if (!(ret=calloc(number,size)))
    fprintf(stderr,"ERROR:CALLOC (%d:%s)",line,file);
  return ret;
}

trfree(char *ptr, int line, char *file)
{
  if (ptr)
    free(ptr);
}
